"""Reproduce the libvroom_example as tests."""
import vroom


def test_example_with_custom_matrix():
    problem_instance = vroom.Input()

    problem_instance.set_durations_matrix(
        profile="car",
        matrix_input=[[0, 2104, 197, 1299],
                      [2103, 0, 2255, 3152],
                      [197, 2256, 0, 1102],
                      [1299, 3153, 1102, 0]],
    )

    problem_instance.add_vehicle(vroom.Vehicle(0, start=0, end=3))
    problem_instance.add_job(vroom.Job(id=1414, location=1))
    problem_instance.add_job(vroom.Job(id=1515, location=2))

    solution = problem_instance.solve(5, 4)

    assert solution.summary.cost == 5461
    assert solution.summary.unassigned == 0
    assert solution.unassigned == []

    [route] = solution.routes
    assert route.vehicle == 0
    assert route.cost == 5461
    assert route.duration == 5461
    assert route.service == 0
    assert route.distance == 0

    assert ([step.step_type for step in route.steps] ==
            [vroom.STEP_TYPE.START, vroom.STEP_TYPE.JOB,
             vroom.STEP_TYPE.JOB, vroom.STEP_TYPE.END])
    assert [step.arrival for step in route.steps] == [0, 2104, 4359, 5461]
    assert [step.duration for step in route.steps] == [0, 2104, 4359, 5461]
    assert [step.service for step in route.steps] == [0, 0, 0, 0]



"""VROOM input definition."""
from numpy.typing import ArrayLike
from typing import Dict, Optional, Union

import numpy

from _vroom import _Input, Matrix, ROUTER, Server


class Input(_Input):
    """VROOM input defintion.

    Main instance for adding jobs, shipments, vehicles, and cost and duration
    matrice defining a routing problem. Duration matrices is if not provided
    can also be retrieved from a map server.

    Attributes:
        jobs:
            Jobs that needs to be completed in the routing problem.
        vehicles:
            Vehicles available to solve the routing problem.

    """

    def __init__(
        self,
        amount_size: int = 0,
        servers: Optional[Dict[str, Union[str, Server]]] = None,
        router: ROUTER = ROUTER.OSRM,
    ) -> None:
        """Class initializer.

        Args:
            amount_size:
                The size of the job to be transported. Used to verify all jobs
                have the same size limit.
            server:
                Assuming no custom duration matrix is provided (from
                `set_durations_matrix`), use coordinates and a map server to
                calculate durations matrix. Keys should be identifed by
                `add_routing_wrapper`. If string, values should be on the
                format `{host}:{port}`.
            router:
                If servers is used, define what kind of server is provided.
                See `vroom.ROUTER` enum for options.
        """
        if servers is None:
            servers = {}
        for key, server in servers.items():
            if isinstance(server, str):
                servers[key] = Server(*server.split(":"))
        self._amount_size = amount_size
        self._servers = servers
        self._router = router
        _Input.__init__(
            self,
            amount_size=amount_size,
            servers=servers,
            router=router,
        )

    def __repr__(self) -> str:
        """String representation."""
        args = []
        if self._amount_size:
            args.append(f"amount_size={self._amount_size}")
        if self._servers:
            args.append(f"servers={self._servers}")
        if self._router != ROUTER.OSRM:
            args.append(f"router={self._router}")
        return f"{self.__class__.__name__}({', '.join(args)})"

    def set_durations_matrix(
            self,
            profile: str,
            matrix_input: ArrayLike,
    ) -> None:
        """Set durations matrix.

        Args:
            profile:
                Name of the transportation category profile in question.
                Typically "car", "truck", etc.
            matrix_input:
                A square matrix consisting of duration between each location of
                interest. Diagonal is canonically set to 0.
        """
        assert isinstance(profile, str)
        if not isinstance(matrix_input, Matrix):
            matrix_input = Matrix(numpy.asarray(matrix_input, dtype="uint32"))
        _Input.set_durations_matrix(self, profile, matrix_input)



from typing import Optional, Sequence, Tuple, Union
from _vroom import _Location


class Location(_Location):
    """Location for where a job needs to e done.

    Either as an index referring to a column in the durations
    matrix, or as longitude-latitude coordinates.

    Attributes:
        index:
            Location index referring to column in the duration
            matrix.
        coords:
            Longitude and latitude coordinate.

    Examples:
        >>> vroom.Location(index=4)
        vroom.Location(index=4)
        >>> loc = vroom.Location(coords=(7., 8.))
        >>> loc.coords
        (7.0, 8.0)

    """

    def __init__(
            self,
            index: Optional[int] = None,
            coords: Optional[Sequence[float]] = None,
    ) -> None:
        if index is None and coords is None:
            raise KeyError(
                "either `index` or `coords` (or both) must be provided.")

        kwargs = {}
        if index is not None:
            assert isinstance(index, int)
            kwargs["index"] = index
        if coords is not None:
            assert isinstance(coords, Sequence)
            assert len(coords) == 2
            kwargs["coords"] = tuple(float(coord) for coord in coords)
        self._kwargs = kwargs
        _Location.__init__(self, **kwargs)

    def __repr__(self):
        args = ", ".join(f"{key}={value}"
                         for key, value in self._kwargs.items())
        return f"vroom.{self.__class__.__name__}({args})"

    @staticmethod
    def from_args(
            args: Union["Location", int, Sequence[Union[int, float]]],
    ) -> "Location":
        """Convenience constructor.

        Allows for short-hand construction from type and length.

        -----------  ---------------------------------------------
        condition    initializer
        -----------  ---------------------------------------------
        Location     index=args.index, coords=[args.lat, args.lon]
        integer      index=args
        length == 1  index=args[0]
        length == 2  coords=args
        length == 3  index=args[0], coords=args[1:]
        -----------  ---------------------------------------------

        Args:
            args:
                Input to interpret as index, coordinate or
                (index, coordinate) pair.

        Examples:
            >>> vroom.Location.from_args(1)
            vroom.Location(index=1)
            >>> vroom.Location.from_args([3., 4.])
            vroom.Location(coords=(3.0, 4.0))
            >>> vroom.Location.from_args([1, 3., 4.])
            vroom.Location(index=1, coords=(3.0, 4.0))
            >>> vroom.Location.from_args(Location(1, coords=[3., 5.]))
            vroom.Location(index=1, coords=(3.0, 5.0))
        """
        if isinstance(args, _Location):
            kwargs = dict(index=args._index())
            if args.has_coordinates():
                kwargs["coords"] = [args.lon(), args.lat()]
            return Location(**kwargs)
        if isinstance(args, int):
            return Location(index=args)
        assert 0 < len(args) < 4
        if len(args) == 1:
            return Location(index=args[0])
        if len(args) == 2:
            return Location(coords=args)
        return Location(index=args[0], coords=args[1:])

    @property
    def index(self) -> Optional[int]:
        """Location index.

        Examples:
            >>> loc = vroom.Location(4)
            >>> loc.index
            4
            >>> loc.index = 7
            >>> loc.index
            7
        """
        return self._index()

    @index.setter
    def index(self, value: int) -> None:
        assert isinstance(value, int)
        self.set_index(value)

    @property
    def coords(self) -> Optional[Tuple[float, float]]:
        """Location longitude and latitude coordinates.

        Examples:
            >>> vroom.Location(coords=[4, 5]).coords
            (4.0, 5.0)
            >>> vroom.Location(index=3).coords is None
            True

        """
        if self.has_coordinates():
            return self.lon(), self.lat()
        return None



from typing import Optional, Sequence, Union

from _vroom import _VehicleStep, JOB_TYPE, STEP_TYPE

from .forced_service import ForcedService


class VehicleStep(_VehicleStep):

    def __init__(
        self,
        *,
        id: Optional[int] = None,
        type: Optional[STEP_TYPE] = None,
        job_type: Optional[JOB_TYPE] = None,
        forced_service: Union[None, Sequence[int], ForcedService] = None,
    ) -> None:
        kwargs = dict(
            id=id,
            type=type,
            job_type=job_type,
            forced_service=forced_service,
        )
        kwargs = {key: value for key, value in kwargs.items()
                  if value is not None}
        self._kwargs = kwargs.copy()
        if isinstance(forced_service, Sequence):
            kwargs["forced_service"] = ForcedService(*forced_service)
        _VehicleStep(**kwargs)

    def __repr__(self) -> str:
        args = [f"{key}={value!r}" for key, value in self._kwargs.items()]
        return f"{self.__class__.__name__}({args})"



"""Time window for when a delivery/pickup is possible."""
from typing import Optional, Sequence, Union
from _vroom import _TimeWindow


class TimeWindow(_TimeWindow):
    """Time window for when a delivery/pickup is possible.

    Attributes:
        start:
            Start point (inclusice) of the time window.
        end:
            End point (inclusive) of the time window.
        length:
            Length of the time interval.

    Examples:
        >>> tw = vroom.TimeWindow(4, 6)
        >>> tw
        vroom.TimeWindow(4, 6)
        >>> tw.start, tw.end, tw.length
        (4, 6, 2)
        >>> 10 in tw, 5 in tw
        (False, True)
    """

    def __init__(
        self,
        start: int = 0,
        end: Optional[int] = None,
    ) -> None:
        """Class initializer.

        Args:
            start:
                Start point (inclusice) of the time window. In
                seconds from the starting time.
            end:
                End point (inclusive) of the time window. In
                seconds from the starting time. If included, have `start < end`
        """
        kwargs = {"start": int(start)}
        if end is not None:
            assert start < end
            kwargs["end"] = int(end)
        _TimeWindow.__init__(self, **kwargs)

    @staticmethod
    def from_args(
        args: Union["TimeWindow", int, Sequence[int]],
    ) -> "TimeWindow":
        """Convenience constructor.

        Allows for short-hand construction.

        -----------  ------------------------------
        condition    initializer
        -----------  ------------------------------
        TimeWindow   start=args.start, end=args.end
        integer      start=args
        length == 2  start=args[0], end=args[1]
        -----------  ------------------------------

        Args:
            args:
                Input to interpret as start, end or TimeWindow pair.

        Examples:
            >>> vroom.TimeWindow.from_args([8, 16])
            vroom.TimeWindow(8, 16)
            >>> vroom.TimeWindow.from_args(TimeWindow(8, 16))
            vroom.TimeWindow(8, 16)

        """
        if isinstance(args, _TimeWindow):
            return TimeWindow(args.start, args.end)
        assert len(args) == 2
        return TimeWindow(*args)

    def __repr__(self):
        return f"vroom.{self.__class__.__name__}({self.start}, {self.end})"



from _vroom import _Amount


class Amount(_Amount):

    def __init__(self, size: int = 0) -> None:
        _Amount.__init__(self, size=size)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.size})"

    @property
    def size(self) -> int:
        return self._size()




"""Global configuration."""
import pytest
import vroom


@pytest.fixture(autouse=True)
def global_setup(doctest_namespace, monkeypatch):
    """Global configuration setup."""
    doctest_namespace["vroom"] = vroom



"""Vehicle routing open-source optimization machine (VROOM)."""
from .break_ import Break
from .job import Job, JOB_TYPE
from .location import Location
from .time_window import TimeWindow
from .vehicle import Vehicle

from .input.forced_service import ForcedService
from .input.input import Input
from .input.vehicle_step import VehicleStep, STEP_TYPE



from typing import List, Optional, Sequence, Set, Union

from _vroom import _Job, JOB_TYPE

from .amount import Amount
from .location import Location
from .time_window import TimeWindow


class Job(_Job):
    """A job with deliver and/or pickup that has to be performed.

    Examples:
        >>> vroom.Job(0, [4., 5.], delivery=4, pickup=7)
        vroom.Job(0, [4.0, 5.0], delivery=4, pickup=7)
    """
    def __init__(
        self,
        id: int,
        location: Union[Location, int, Sequence[Union[int, float]]],
        type: JOB_TYPE = JOB_TYPE.SINGLE,
        setup: int = 0,
        service: int = 0,
        delivery: Union[int, Amount] = 0,
        pickup: Union[int, Amount] = 0,
        skills: Optional[Set[int]] = None,
        priority: int = 0,
        time_windows: Optional[Sequence[TimeWindow]] = None,
        description: str = "",
    ) -> None:
        """
        Args:
            id:
                Job identifier number. Two jobs can not have the same
                identifier.
            type:
                The type of job that has to be performed.
            location:
                Location of the job. If iterger, value interpreted as an the
                column in duration matrix. If pair of floats, value interpreted
                as longitude and latitude coordinates respectively. If a
                triplet, it is interpreted as `(index, longitude, latitude)`.
            setup:
                The cost of preparing the vehicle before actually going out for
                a job.
            service:
                The time (in secondes) it takes to pick up/deliver shipment
                when at customer.
            delivery:
                An interger representation of how much is being carried to
                customer.
            pickup:
                An interger representation of how much is being carried back
                from customer.
            skills:
                Skills required to perform job. Only vehicles which satisfies
                all required skills (i.e. has at minimum all skills values
                required) are allowed to perform this job.
            priority:
                The job priority level, where 0 is the most
                important and 100 is the least important.
            time_windows:
                Windows for where service is allowed to begin.
                Defaults to have not restraints.
            description:
                Optional string descriping the job.
        """
        assert isinstance(id, int)
        assert isinstance(pickup, int)
        assert isinstance(delivery, int)
        kwargs = dict(
            id=id,
            location=location,
            type=type,
            setup=setup,
            service=service,
            delivery=delivery,
            pickup=pickup,
            skills=skills,
            priority=priority,
            time_windows=time_windows,
            description=description,
        )
        kwargs = {key: value for key, value in kwargs.items()
                  if value or key == "id"}
        self._kwargs = kwargs.copy()
        kwargs["location"] = Location.from_args(location)
        if time_windows is not None:
            kwargs["tws"] = [TimeWindow.from_args(tw)
                             for tw in kwargs.pop("time_windows")]
        assert isinstance(type, JOB_TYPE)
        if type == JOB_TYPE.SINGLE:
            kwargs["delivery"] = Amount(delivery)
            kwargs["pickup"] = Amount(pickup)
            del kwargs["type"]
            del self._kwargs["type"]
        elif type == JOB_TYPE.DELIVERY:
            kwargs["delivery"] = Amount(delivery)
            assert "pickup" not in kwargs
        elif type == JOB_TYPE.PICKUP:
            kwargs["pickup"] = Amount(pickup)
            assert "delivery" not in kwargs

        _Job.__init__(self, **kwargs)

    def __repr__(self) -> str:
        kwargs = {key: value for key, value in self._kwargs.items()
                  if value or key == "id"}
        id = kwargs.pop("id")
        location = kwargs.pop("location")
        args = ", ".join(f"{key}={value}" for key, value in kwargs.items())
        return f"vroom.{self.__class__.__name__}({id}, {location}, {args})"

    @property
    def location(self) -> Location:
        """
        The location where to go.

        Either by index (used with duration matrix) or
        by coordinate (used with map server).
        """
        return Location.from_args(self._kwargs["location"])

    @property
    def time_windows(self) -> List[TimeWindow]:
        """Time window for when job can be delivered."""
        return [TimeWindow.from_args(tw) for tw in self._kwargs.get("time_windows", [])]



from _vroom import _Vehicle
from typing import List, Optional, Sequence, Set, Union

from .amount import Amount
from .break_ import Break
from .location import Location
from .time_window import TimeWindow
from .input.vehicle_step import VehicleStep


class Vehicle(_Vehicle):
    """Vehicle for performing transport.

    Examples:
        >>> vehicle = vroom.Vehicle(1, end=1)
        >>> vehicle
        vroom.Vehicle(1, end=1)
    """

    def __init__(
        self,
        id: int,
        start: Union[None, Location, int, Sequence[Union[int, float]]] = None,
        end: Union[None, Location, int, Sequence[Union[int, float]]] = None,
        profile: Optional[str] = None,
        capacity: int = 0,
        skills: Optional[Set[int]] = None,
        time_window: Optional[TimeWindow] = None,
        breaks: Sequence[Break] = (),
        description: str = "",
        speed_factor: Optional[float] = None,
        max_tasks: Optional[int] = None,
        input_steps: Sequence[VehicleStep] = (),
    ) -> None:
        kwargs = dict(
            id=id,
            start=start,
            end=end,
            profile=profile,
            capacity=capacity,
            skills=skills,
            time_window=time_window,
            breaks=breaks,
            description=description,
            speed_factor=speed_factor,
            max_tasks=max_tasks,
            input_steps=input_steps,
        )
        kwargs = {key: value for key, value in kwargs.items()
                  if value or key == "id"}
        self._kwargs = kwargs.copy()
        kwargs["capacity"] = Amount(capacity)
        kwargs["start"] = None if start is None else Location.from_args(start)
        kwargs["end"] = None if end is None else Location.from_args(end)
        if "time_window" in kwargs:
            kwargs["tw"] = kwargs.pop("time_window")
        _Vehicle.__init__(self, **kwargs)

    def __repr__(self) -> str:
        kwargs = {key: value for key, value in self._kwargs.items()
                  if key == "id" or value}
        id = kwargs.pop("id")
        args = ", ".join(f"{key}={value!r}" for key, value in kwargs.items())
        return f"vroom.{self.__class__.__name__}({id}, {args})"

    @property
    def start(self) -> Optional[Location]:
        return Location.from_args(self._start) if self.has_start() else None

    @property
    def end(self) -> Optional[Location]:
        return Location.from_args(self._end) if self.has_end() else None

    @property
    def time_window(self) -> TimeWindow:
        return TimeWindow.from_args(self.tw)

    @property
    def input_steps(self) -> List[VehicleStep]:
        return [VehicleStep.from_args(step) for step in self.steps]



import os
from setuptools import setup
from pybind11.setup_helpers import Pybind11Extension, build_ext

ext_modules = [
    Pybind11Extension(
        "_vroom",
        [os.path.join("src", "_vroom.cpp")],
        extra_compile_args=[
            "-MMD",
            "-MP",
            "-Wextra",
            "-Wpedantic",
            "-Wall",
            "-O3",
            "-DASIO_STANDALONE",
            "-DNDEBUG",
        ],
        extra_link_args=[
            "-lpthread",
            "-lssl",
            "-lcrypto",
        ],
    ),
]

setup(
    cmdclass={"build_ext": build_ext},
    ext_modules=ext_modules,
    include_dirs=[os.path.join("vroom", "src")],
    use_scm_version = True,
)




from _vroom import _ForcedService
from typing import Optional

class ForcedService(_ForcedService):
    """
    Examples:
        >>> vroom.ForcedService()
        vroom.ForcedService()
    """

    def __init__(
        self,
        at: Optional[int] = None,
        after: Optional[int] = None,
        before: Optional[int] = None,
    ) -> None:
        kwargs = dict(at=at, after=after, before=before)
        self._kwargs = {key: value for key, value in kwargs.items()
                        if value is not None}
        _ForcedService.__init__(self, **self._kwargs)

    def __repr__(self) -> str:
        args = ", ".join(f"{key}={value}"
                         for key, value in self._kwargs.items())
        return f"vroom.{self.__class__.__name__}({args})"

    @property
    def at(self) -> int:
        return self._kwargs["at"]

    @property
    def after(self) -> int:
        return self._kwargs["after"]

    @property
    def before(self) -> int:
        return self._kwargs["before"]
